/*
+----------------------------------------------------------------------+
|                    Class: CLauncherRestConnectionManager             |
|                                                                      |
| Developper:  Eric Gavaldo (eric.gavaldo@xqual.com)                   |
+----------------------------------------------------------------------+
*/
package com.xqual.xlauncher.mrsuricate;

import java.io.BufferedWriter;
import java.io.File;
import java.io.OutputStreamWriter;
import java.net.HttpURLConnection;
import java.net.URL;

import javax.net.ssl.HostnameVerifier;
import javax.net.ssl.HttpsURLConnection;
import javax.net.ssl.SSLContext;
import javax.net.ssl.SSLSocketFactory;
import javax.net.ssl.TrustManager;

import org.json.JSONObject;

import com.xqual.xcommon.utils.CFileUtils;
import com.xqual.xcommon.utils.CTextUtils;
import com.xqual.xcommon.utils.network.CYesHostnameVerifier;
import com.xqual.xcommon.utils.network.CYesTrustManager;

public class CLauncherRestConnectionManager {

	// +==============================================================+
	// | Attributes                                                   |
	// +==============================================================+
	
	private final String baseUrl;
	private final String email;
	private final String password;

	// returned by the API
	public String userId;
	private String bearerToken;
	
	private final boolean httpsEnabled;
	
	// +==============================================================+
	// | Constructors                                                 |
	// +==============================================================+
	
	public CLauncherRestConnectionManager(String baseUrl, String email, String password) {
	    this.baseUrl  = baseUrl;   // i.e. "https://api.mrsuricate.com/api-v2"
		this.email    = email;     // i.e. "my_email@acme.com"
		this.password = password;  // i.e. "my_password"

	    HostnameVerifier yesHostnameVerifier = new CYesHostnameVerifier();
	    TrustManager yesTrustManager = new CYesTrustManager();
        
        URL url = null;
	    try {
	    	url = new URL(baseUrl);
		} catch (Throwable t) {
			t.printStackTrace();
			httpsEnabled= false;
			return;
		}
	    
	    // ------------------------------
	    // configure the HTTPS connection
	    // ------------------------------
	    if (url.getProtocol().equalsIgnoreCase("https")) {
			System.out.println("the Mr Suricate server is an https/SSL server");
		    httpsEnabled = true;
		    
		    // USE THE SAME SETTINGS THAN FOR DATABASE (database.connection.accept_any_certificate)
		    //if (CStudioGlobals.applicationConfProperties.isAcceptAnyCertificate()) { // the user accepts ANY certificate
		    	System.out.println("==> accepting ANY certificate");
		        // initiate the SSL context and prepare trust manager that will accept ANY certificate
		        TrustManager[] yesTrustManagers = new TrustManager[] {yesTrustManager};
		        try {
					SSLContext yesSslcontext = SSLContext.getInstance("TLS");
					yesSslcontext.init(null, yesTrustManagers, null); // key manager, trust manager, secure random manager
					SSLSocketFactory yesSslFactory = yesSslcontext.getSocketFactory();
					
				    HttpsURLConnection.setDefaultSSLSocketFactory(yesSslFactory);
				    HttpsURLConnection.setDefaultHostnameVerifier(yesHostnameVerifier);
		        } catch (Exception e) {
				}
		        
		    /*} else {
		    	System.out.println(" ==> accepting only certificate matching a certificate from the local/private keystore");
		    	HttpsURLConnection.setDefaultSSLSocketFactory(CStudioGlobals.xstudioSslFactory); // this will trust the server only if its certificates has been added in the conf
		    }*/

	    // -----------------------------
	    // configure the HTTP connection
	    // -----------------------------
	    } else {
	    	System.out.println("the XStudio server is an http server");
		    httpsEnabled = false;
	    }
	}
	
	// +==============================================================+
	// | Methods                                                      |
	// +==============================================================+

	public void authenticate() {
		System.out.println("============");
		System.out.println("Authenticate");
		System.out.println("============");
		StringBuffer jsonInput = new StringBuffer();
		jsonInput.append("{");
		jsonInput.append("\"email\":\"" + CTextUtils.escapeJsonValue(this.email) + "\",");
		jsonInput.append("\"password\":\"" + CTextUtils.escapeJsonValue(this.password) + "\"");
		jsonInput.append("}");
		
		String jsonContent = postData("/auth/login", jsonInput.toString());
		System.out.println("- - - - - jsonContent = " + jsonContent);
		JSONObject jsonObj = new JSONObject(jsonContent);
		bearerToken = jsonObj.getString("token");
		System.out.println("---------------------------------------------------");
		System.out.println("==> bearer token = " + bearerToken);
		System.out.println("---------------------------------------------------");
	}
	
	public String getData(String urlPostfix) {
		HttpURLConnection connectionGet = null;
		try {
	        // prepare the connection
	        URL url = new URL(baseUrl + urlPostfix);
	        System.out.println("connecting to url \"" + url + "\"");
	        
	        // the connection may already be opened so we catch the IllegalStateException "Already Connected"
	        try {
				if (httpsEnabled) {
					connectionGet = (HttpsURLConnection)url.openConnection(); // HttpsURLConnection extends HttpURLConnection

				} else if (url.getProtocol().equalsIgnoreCase("http")) {
					connectionGet = (HttpURLConnection)url.openConnection();
				}
	        
	            connectionGet.setRequestMethod("GET");
	            /*
	            http://stackoverflow.com/questions/17278303/basic-authentication-fails-with-glassfish
	            Oracle has enabled caching of HTTP responses by default with JDK7 in Web Start: 
				Caching enabled by default: Caching of network content for application code running in Web Start mode is now enabled by default. 
				This allows application improved performance and consistency with applet execution mode. 
				To ensure the latest copy of content is used, the application can use URLConnection.setUseCaches(false) or request header Cache-Control values no-cache/no-store.
				*/
	            // maybe we should try to re-enable cache after the first connection?
		        connectionGet.setRequestProperty("Cache-Control", "no-cache");
		        connectionGet.setRequestProperty("Cache-Control", "no-store");
		        if (bearerToken != null)
		        	connectionGet.setRequestProperty("Authorization", "Bearer " + bearerToken);
		        connectionGet.setRequestProperty("Connection", "keep-alive");
		        
	        } catch (IllegalStateException ise) {
				System.out.println("Already connected\n");
			}
	        
	        // ----------- CONNECTION -----------
	        connectionGet.connect(); // that's where security handshake is done

	        // store the code and the message so that the caller can retrieve this
	        int responseCode = connectionGet.getResponseCode();
	        String responseMessage = connectionGet.getResponseMessage();

	        if (responseCode != 200) {
	        	System.err.println("Server response: " + responseCode + " (" + responseMessage + ")");
	        } else {
	        	//System.out.println("Server response: " + responseCode + " (" + responseMessage + ")");
	        }
	        
	        if (responseCode != 200) {
	        	String output = CFileUtils.streamToString(connectionGet.getErrorStream());
		        System.err.println("ERROR=" + output);
				try {
					System.err.println("Disconnecting...");
					connectionGet.disconnect();
				} catch (Exception e) {
					System.err.println("Couldn't disconnect...");
				}
				System.err.println("returning null...");
	        	return null;
	        }
	        
	        String output = CFileUtils.streamToString(connectionGet.getInputStream());
	        //System.err.println("OUTPUT=" + output);
	        connectionGet.disconnect();
	        return output;
	        
		} catch (Throwable t) {
			System.err.println("Exception while communicating with VTaaS REST server: " + t);
			try {
				System.err.println("Disconnecting...");
				connectionGet.disconnect();
			} catch (Exception e) {
				System.err.println("Couldn't disconnect...");
			}
			return null;
		}
	}	


	public File downloadData(String urlPostfix) {
		HttpURLConnection connectionGet = null;
		try {
	        // prepare the connection
	        URL url = new URL(baseUrl + urlPostfix);
	        System.out.println("connecting to url \"" + url + "\"");
	        
	        // the connection may already be opened so we catch the IllegalStateException "Already Connected"
	        try {
				if (httpsEnabled) {
					connectionGet = (HttpsURLConnection)url.openConnection(); // HttpsURLConnection extends HttpURLConnection

				} else if (url.getProtocol().equalsIgnoreCase("http")) {
					connectionGet = (HttpURLConnection)url.openConnection();
				}
	        
	            connectionGet.setRequestMethod("POST");
	            // maybe we should try to re-enable cache after the first connection?
		        connectionGet.setRequestProperty("Cache-Control", "no-cache");
		        connectionGet.setRequestProperty("Cache-Control", "no-store");
		        if (bearerToken != null)
		        	connectionGet.setRequestProperty("Authorization", "Bearer " + bearerToken);
		        connectionGet.setRequestProperty("Connection", "keep-alive");
		        
	        } catch (IllegalStateException ise) {
				System.out.println("Already connected\n");
			}
	        
	        // ----------- CONNECTION -----------
	        connectionGet.connect(); // that's where security handshake is done
	        //if (httpsEnabled) {
	        //    CStudioSqUtils.printHttpsCertificateChain((HttpsURLConnection)connection); //dump all cert info
	        //}

	        // store the code and the message so that the caller can retrieve this
	        int responseCode = connectionGet.getResponseCode();
	        String responseMessage = connectionGet.getResponseMessage();

	        if (responseCode != 200) {
	        	System.err.println("Server response: " + responseCode + " (" + responseMessage + ")");
	        } else {
	        	//System.out.println("Server response: " + responseCode + " (" + responseMessage + ")");
	        }
	        
	        if (responseCode != 200) {
	        	String output = CFileUtils.streamToString(connectionGet.getErrorStream());
		        System.err.println("ERROR=" + output);
				try {
					System.err.println("Disconnecting...");
					connectionGet.disconnect();
				} catch (Exception e) {
					System.err.println("Couldn't disconnect...");
				}
	        	return null;
	        }
	        
	        File outputFile = CFileUtils.createTmpFile("screenshot.png");
	        CFileUtils.streamToFile(connectionGet.getInputStream(), outputFile);
	        //System.err.println("OUTPUT=" + output);
	        connectionGet.disconnect();
	        return outputFile;
	        
		} catch (Throwable t) {
			System.err.println("Exception while communicating with XStudio REST server: " + t);
			try {
				System.err.println("Disconnecting...");
				connectionGet.disconnect();
			} catch (Exception e) {
				System.err.println("Couldn't disconnect...");
			}
			return null;
		}
	}	
	
	public String postData(String urlPostfix, String jsonData) {
		HttpURLConnection connectionPost = null;
		try {
	        // prepare the connection
	        URL url = new URL(baseUrl + urlPostfix);
	        System.out.println("connecting to url \"" + url + "\"");
	        
	        // the connection may already be opened so we catch the IllegalStateException "Already Connected"
	        try {
				if (httpsEnabled) {
					connectionPost = (HttpsURLConnection)url.openConnection(); // HttpsURLConnection extends HttpURLConnection

				} else if (url.getProtocol().equalsIgnoreCase("http")) {
					connectionPost = (HttpURLConnection)url.openConnection();
				}
	        
	            connectionPost.setRequestMethod("POST");
	            /*
	            http://stackoverflow.com/questions/17278303/basic-authentication-fails-with-glassfish
	            Oracle has enabled caching of HTTP responses by default with JDK7 in Web Start: 
				Caching enabled by default: Caching of network content for application code running in Web Start mode is now enabled by default. 
				This allows application improved performance and consistency with applet execution mode. 
				To ensure the latest copy of content is used, the application can use URLConnection.setUseCaches(false) or request header Cache-Control values no-cache/no-store.
				*/
	            // maybe we should try to re-enable cache after the first connection?
		        connectionPost.setRequestProperty("Cache-Control", "no-cache");
		        connectionPost.setRequestProperty("Cache-Control", "no-store");
		        if (bearerToken != null)
		        	connectionPost.setRequestProperty("Authorization", "Bearer " + bearerToken);
		        connectionPost.setRequestProperty("Connection", "keep-alive");
		        
		        // format of the body
		        connectionPost.setRequestProperty("Content-Type", "application/json;charset=utf-8");

				// Send post request
		        connectionPost.setDoInput(true);
		        connectionPost.setDoOutput(true);
				BufferedWriter bw = new BufferedWriter(new OutputStreamWriter(connectionPost.getOutputStream(), "UTF-8"));
				
				bw.write(jsonData);
				bw.flush();
				bw.close();
	        } catch (IllegalStateException ise) {
				System.out.println("Already connected\n");
			}
	        
	        
	        // ----------- CONNECTION -----------
	        connectionPost.connect(); // that's where security handshake is done

	        // store the code and the message so that the caller can retrieve this
	        int responseCode = connectionPost.getResponseCode();
	        String responseMessage = connectionPost.getResponseMessage();

	        if (responseCode != 200 && responseCode != 201) { // 201=Created
	        	System.err.println("Server response: " + responseCode + " (" + responseMessage + ")");
	        } else {
	        	//System.out.println("Server response: " + responseCode + " (" + responseMessage + ")");
	        }
	        
	        if (responseCode != 200 && responseCode != 201) { // 201=Created
	        	//String output = CFileUtils.streamToString(connectionPost.getErrorStream());
		        //System.err.println("ERROR=" + output);
				try {
					System.err.println("Disconnecting...");
					connectionPost.disconnect();
				} catch (Exception e) {
					System.err.println("Couldn't disconnect...");
				}
				System.err.println("returning null...");
	        	return null;
	        }
	        
	        String output = CFileUtils.streamToString(connectionPost.getInputStream());
	        //System.err.println("OUTPUT=" + output);
	        connectionPost.disconnect();
	        return output;
	        
		} catch (Throwable t) {
			System.err.println("Exception while communicating with XStudio REST server: " + t);
			try {
				System.err.println("Disconnecting...");
				connectionPost.disconnect();
			} catch (Exception e) {
				System.err.println("Couldn't disconnect...");
			}
			return null;
		}
	}
	
	public String getBaseUrl() {
		return baseUrl;
	}

}
